import sqlalchemy
from sqlalchemy import orm, Column, Integer, String, Enum, ForeignKey
from sqlalchemy.ext import declarative
import os

sql_alchemy_base = declarative.declarative_base()
object_types_enum = Enum('champion', 'item', 'rune')


class LolObject(sql_alchemy_base):
    """Represents a LoL-related object with its ID, name, locale, and object type.
    """
    __tablename__ = 'lol_id'

    # The id from Riot
    id = Column(Integer, primary_key=True)

    # The locale the name is in
    locale = Column(String, primary_key=True)

    # The name in the given locale
    name = Column(String)

    # The object type, might need a change in the future
    object_type = Column(object_types_enum)

    def __init__(self, id_, locale_, name_, object_type_):
        self.id = id_
        self.locale = locale_
        self.name = name_
        self.object_type = object_type_


class GhostSession:
    """A ghost loader for our sqlalchemy session

    This makes it sure we do not create the session and schema during module import.
    """
    session = None

    @classmethod
    def get_session(cls):
        if not cls.session:
            data_folder = os.path.join(os.path.expanduser("~"), '.config', 'lol_id_tools')
            if not os.path.exists(data_folder):
                os.mkdir(data_folder)

            engine = sqlalchemy.create_engine(f'sqlite:///{os.path.join(data_folder, "lol_id_tools.db")}')
            sql_alchemy_base.metadata.create_all(engine)
            cls.session = orm.sessionmaker(bind=engine)()
        return cls.session


# The ghost session function
ghost_session = GhostSession.get_session

##




import asyncio
import rapidfuzz
from lol_id_tools.data.local_data import nicknames, clean_locale
from lol_id_tools.sqlite_classes import ghost_session, LolObject
from lol_id_tools.loaders import locale_in_database, load_locale, reload_data


def get_name(id_: int, locale: str = 'en_US', retry=False) -> str:
    """Gets you the name of the associated Riot object.

    If the chosen locale is not loaded in the database, it will be loaded before returning the result.

    Args:
        id_: Riot ID of the object.
        locale: Locale of the output.
        retry: Optional variable specifying if data should be reloaded if the object cannot be found.

    Returns:
        The matching object name.

    Raises:
        KeyError: The corresponding object was not found.
        ValueError: The locale was not understood properly.

    Usage example:
        get_name(21)
        get_name(21, 'ko_KR')
    """
    # We start by cleaning up our locale
    locale = clean_locale(locale)

    # First, we see if the object is there with the given constraints
    # Not using walrus operator to keep 3.5+ compatibility.
    try:
        return ghost_session().query(LolObject)\
            .filter(LolObject.id == id_, LolObject.locale == locale).one_or_none().name
    except AttributeError:
        pass

    if not locale_in_database(locale):
        asyncio.run(load_locale(locale))
        return get_name(id_, locale, False)

    if retry:
        asyncio.run(reload_data())
        return get_name(id_, locale, False)

    raise KeyError('The associated Riot object could not be found.')


class NoMatchingNameFound(Exception):
    pass


def get_id(object_name: str, minimum_score: int = 75,
           input_locale: str = None, object_type: str = None, retry: bool = False) -> int:
    """Returns the best Riot ID guess for the given name.

    Will return the id that matches the query best. Best used in curated environments.

    Args:
        object_name: Search string.
        minimum_score: Optional minimum Levensthein distance score under which the function retries or raises.
        input_locale: The language the input was in.
        object_type: Optional boolean specifying the type of the object in [champion, item, rune].
        retry: Optional variable specifying if data should be reloaded if the object cannot be found.

    Returns:
        The matching object ID.

    Raises:
        NoMatchingNameFound: No string scored above the minimum_score.
        ValueError: The locale was not understood properly.

    Usage example:
        get_id('Miss Fortune')
        get_id('MF')
        get_id('미스 포츈')
        get_id('Dio')
    """
    # Handling some Leaguepedia special cases as a 0 item (
    if not object_name or object_name == 'None' or object_name == 'Loss of Ban' or object_name == 'No item':
        return 0

    # First, we try to directly get the object with the exact input name
    try:
        return ghost_session().query(LolObject.id).filter(LolObject.name == object_name).first().id
    except AttributeError:
        pass

    names_query = ghost_session().query(LolObject.id, LolObject.name)

    if object_type:
        names_query = names_query.filter(LolObject.object_type == object_type)

    if input_locale:
        locale = clean_locale(input_locale)

        # If the locale was not loaded yet, we restart the process
        if not locale_in_database(locale):
            asyncio.run(load_locale(locale))
            return get_id(object_name, minimum_score, input_locale, object_type, False)

        # If it was, we filter on it
        names_query = names_query.filter(LolObject.locale == locale)

    # TODO Review how we work with nicknames, likely put them in the sqlite DB too
    # TODO Find a smarter way to do the fuzzy matching with sqlite or change backend
    name_guess, score = rapidfuzz.process.extractOne(object_name,
                                                     [row.name for row in names_query] + list(nicknames.keys()))

    # If our result is a nickname, we go get the clean name
    if name_guess in nicknames:
        name_guess = nicknames[name_guess]

    if score < minimum_score:
        if not retry:
            raise NoMatchingNameFound('No object name close enough to the input string found.')
        else:
            reload_data()
            return get_id(object_name, minimum_score, input_locale, object_type, False)

    return ghost_session().query(LolObject.id).filter(LolObject.name == name_guess).first().id


def get_translation(object_name: str, output_locale: str = 'en_US', minimum_score: int = 75,
                    input_locale: str = None, object_type: str = None, retry: bool = False) -> str:
    """Returns the best translation guess for the given name.

    Will return the id that matches the query best. Best used in curated environments.

    Args:
        object_name: Search string.
        output_locale: The language to translate to, with 'en_US' as the default value.
        minimum_score: Optional minimum Levensthein distance score under which the function retries or raises.
        input_locale: The language the input was in.
        object_type: Optional boolean specifying the type of the object in [champion, item, rune].
        retry: Optional variable specifying if data should be reloaded if the object cannot be found.

    Returns:
        The matching object ID.

    Raises:
        NoResultFound: No result scored above the minimum_score.
        ValueError: One of the input locales was not understood properly.

    Usage example:
        get_translation('미스 포츈')
        get_translation('Miss Fortune', 'ko_KR')
        get_translation('mf')
    """
    return get_name(get_id(object_name, minimum_score, input_locale, object_type, retry), output_locale)



import json
import os
import rapidfuzz

with open(os.path.join(os.path.dirname(__file__), 'locales.json'), encoding='utf-8') as file:
    locales_dict = json.load(file)

with open(os.path.join(os.path.dirname(__file__), 'nicknames.json'), encoding='utf-8') as file:
    raw_nicknames = json.load(file)
    nicknames = {}
    for locale_ in raw_nicknames:
        nicknames.update(raw_nicknames[locale_])


def clean_locale(locale: str):
    """Returns a "clean" locale from a user-entered input.
    """
    # If not, we check if the locale is "clean". If not, we match it to our language list.
    if locale in locales_dict.values():
        return locale

    matching_language, score = rapidfuzz.process.extractOne(locale, locales_dict.keys())
    if score > 80:
        return locales_dict[matching_language]
    else:
        raise ValueError('The locale name could not be understood')



import logging
from lol_id_tools.sqlite_classes import ghost_session, LolObject

dd_url = 'https://ddragon.leagueoflegends.com'


async def load_objects(http_session, latest_version, locale: str, object_type: str):
    """Loads the selected type of objects in the database.

    Queries the Riot servers for the data then writes it to the database.
    All three object types are meant to be called together.

    Args:
        http_session: ClientSession that issues the query
        latest_version: the latest version available on dd
        locale: the locale to load
        object_type: the type of object to load
    """
    url = get_url(latest_version, locale, object_type)

    async with http_session.get(url) as response:
        logging.debug(f'Querying {url}')
        data = await response.json()

    if object_type == 'champion':
        parse_champions(data, locale)
    elif object_type == 'item':
        parse_items(data, locale)
    elif object_type == 'rune':
        parse_runes(data, locale)

    ghost_session().commit()


def get_url(latest_version, locale: str, object_type: str):
    # Riot changed name for some reason
    if object_type == 'rune':
        object_type = 'runesReforged'

    return f'{dd_url}/cdn/{latest_version}/data/{locale}/{object_type}.json'


def parse_champions(data, locale):
    for champion_tag, champion_dict in data['data'].items():
        lol_object = LolObject(int(champion_dict['key']),
                               locale,
                               champion_dict['name'],
                               'champion')
        ghost_session().merge(lol_object)


def parse_items(data, locale):
    for item_id, item_dict in data['data'].items():
        lol_object = LolObject(int(item_id),
                               locale,
                               item_dict['name'],
                               'item')
        ghost_session().merge(lol_object)


def parse_runes(data, locale):
    for rune_tree in data:
        for slot in rune_tree['slots']:
            for rune in slot['runes']:
                lol_object = LolObject(rune['id'],
                                       locale,
                                       rune['name'],
                                       'rune')
                ghost_session().merge(lol_object)



import asyncio
import logging

import aiohttp
from sqlalchemy import distinct

from lol_id_tools.parsers import load_objects
from lol_id_tools.sqlite_classes import ghost_session, LolObject, object_types_enum


def locale_in_database(locale: str):
    """Checks if the chosen locale has already been loaded.

    Args:
        The locale to check.

    Returns:
        The first object with the corresponding locale or None.
    """
    return ghost_session().query(LolObject.locale).filter(LolObject.locale == locale).first()


async def load_locale(locale, latest_version=None):
    if not latest_version:
        latest_version = await get_latest_version()

    async with aiohttp.ClientSession() as http_session:
        await asyncio.wait([asyncio.create_task(load_objects(http_session, latest_version, locale, object_type))
                            for object_type in object_types_enum.enums])


async def reload_data():
    locales_in_db = ghost_session().query(distinct(LolObject.locale).label('locale'))
    latest_version = await get_latest_version()

    await asyncio.wait([asyncio.create_task(load_locale(row.locale, latest_version)) for row in locales_in_db])


async def get_latest_version():
    async with aiohttp.ClientSession() as http_session:
        url = 'https://ddragon.leagueoflegends.com/api/versions.json'
        async with http_session.get('https://ddragon.leagueoflegends.com/api/versions.json') as response:
            logging.debug(f'Querying {url}')
            data = await response.json()
        return data[0]


def reset_data():
    """Deletes everything in the locale database.

    Mostly for testing purposes.
    """
    lol_objects = ghost_session().query(LolObject)

    for lol_object in lol_objects:
        ghost_session().delete(lol_object)

    ghost_session().commit()
